from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.communicator import Communicator
from iqmotion.custom_errors import IqModuleError

# This is a IQ Module Wrapper for the vertiq 8108
# This Wrapper takes in firmware argument and creates an veritq8108 API Object
class Vertiq8108(IqModule):
    """ Creates vertiq8108 object with default speed firmware
    
    Arguments:
        com {Communicator} -- The communicator object to interface with the IqModule
        firmware {str} -- Which firmware is the motor using (default: {"speed"})
    
    Keyword Arguments:
        module_idn {int} -- The idn of the module (default: {0})
        extra_clients {list} -- list of file paths to extra clients you want to load in the module (default: {None})
    """

    def __init__(
        self,
        com: Communicator, 
        module_idn: int = 0, 
        firmware: str = "speed", # Default Firmware       
        clients_path: str = None
    ):
        if firmware.lower() == "speed":
            self._DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"
            self._MODULE_FILE_NAME = "speed8108.json"

        else:
            raise IqModuleError("'" + str(firmware) + "' firmware is not supported")

        super().__init__(com, module_idn, clients_path)
        


from dataclasses import dataclass
import enum
import struct

from iqmotion.client_entries.client_entry import ClientEntry
from iqmotion.client_entries.client_entry_data import ClientEntryData


class AccessType(enum.Enum):
    GET = 0
    SET = 1
    SAVE = 2
    REPLY = 3


@dataclass
class DictionaryClientEntryData(ClientEntryData):
    type_idn: bytes
    payload_type: bytes
    param_idn: bytes
    format: str
    unit: str
    name: str

    def __init__(self, client_entry_dict: dict):
        self.type_idn = client_entry_dict["type_idn"]
        self.payload_type = 0
        self.param_idn = client_entry_dict["param_idn"]
        self.format = client_entry_dict["format"]
        self.unit = client_entry_dict["unit"]
        self.name = client_entry_dict["param"]

    def __str__(self):
        return "{0:10} | {1:2}: {2:32} {3:4} {4}".format(
            self.type_idn, self.param_idn, self.name, self.format, self.unit
        )


class DictionaryClientEntry(ClientEntry):
    """ DictionaryClientEntry is an implementation of ClientEntry.

        A DictrionaryClientEntry follows this structure:
            {"type_idn":59, "param":"ctrl_angle",
                "param_idn":  3, "format":"f", "unit": "rad" }
    """

    def __init__(self, client_entry_data_dict: dict):
        self._fresh = 0
        self._value = None
        self._data = DictionaryClientEntryData(client_entry_data_dict)

    def read_message(self, msg):
        """ Takes in a message, parses it and save the payload as its value

        General Message Format:
            | type_idn | param_idn | obj/access | value |
            'type_idn' is the (uint8) type identifier
            'param_idn' is the (uint8) param identifier
            'obj/access' high 6 bits are the object identifier, low 2 bits are access direction:
            'value' is the (format) value of the message
        """
        msg_type_idn = msg[0]
        msg_param_idn = msg[1]
        msg_access_type = msg[2] & 3
        msg_value = msg[3:]

        if (msg_type_idn == self.data.type_idn) & (
            msg_param_idn == self.data.param_idn
        ):
            if msg_access_type == AccessType.REPLY.value:
                self.value = msg_value

    @property
    def fresh(self):
        """ Checks if the dictionary client entry value is "fresh" and never been read before

        Returns:
            true:  If value is fresh
            false: If value has been read before
        """
        return self._fresh

    @property
    def value(self):
        """ Gets the value of the Client Entry, the type is define by the format in the client entry

        Returns:
            value (format): Type is defined by format
        """
        self._fresh = 0
        return self._value

    @value.setter
    def value(self, value: bytearray):
        """ Sets the value of the Client Entry and formats it to the right type
        """
        # unpack always returns a tuple
        data_format = self._data.format
        formated_value = struct.unpack(data_format, value)
        if len(data_format) < 2:
            formated_value = formated_value[0]
        self._value = formated_value
        self._fresh = 1

    @property
    def data(self):
        return self._data



# from iqmotion.client_entries.client_entry import ClientEntry
# from iqmotion.client_entries.client_entry_data import ClientEntryData

# from dataclasses import dataclass


# @dataclass
# class ProcessClientEntryData(ClientEntryData):
#     type_idn: bytes
#     payload_type: bytes
#     format: str
#     unit: str
#     name: str

#     def __init__(self, client_entry_dict: dict):
#         self.type_idn = client_entry_dict["type_idn"]
#         self.payload_type = client_entry_dict["payload_type"]
#         self.format = client_entry_dict["format"]
#         self.unit = client_entry_dict["unit"]
#         self.name = client_entry_dict["param"]

#     def __str__(self):
#         return "{0:10} | {2:32} {3:4} {4}".format(self.type_idn, self.name, self.format, self.unit)


# class ProcessClientEntry(ClientEntry):
#     def __init__(self, client_entry_data_dict: dict):
#         self._fresh = 0
#         self._value = None
#         self._data = ProcessClientEntryData(client_entry_data_dict)

#     def read_message(self, msg):
#         """ Takes in a message, parses it and save the payload as its value
#         """
#         pass

#     @property
#     def fresh(self):
#         return self._fresh

#     @property
#     def value(self):
#         return self._value

#     @value.setter
#     def value(self, value):
#         pass

#     @property
#     def data(self):
#         return self._data



from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.communicator import Communicator


class BaseIqModule(IqModule):
    _MODULE_FILE_NAME = "base.json"

    def __init__(
        self,
        com: Communicator,
        module_idn: int = 0,
        clients_path: str = None,
    ):

        super().__init__(com, module_idn, clients_path)



from abc import ABC, abstractmethod


class Client(ABC):
    """ Client is an interface class used by the different client implementations.
    A Client object is able to read a message to store its content if needed, check if it's fresh and retrieve it
    """

    @abstractmethod
    def read_message(self, message: bytearray):
        """ this property is too abstract to understand. """

    @abstractmethod
    def is_fresh(self, value_name: str = ""):
        """ this property is too abstract to understand. """

    @abstractmethod
    def get_reply(self, value_name: str = ""):
        """ this property is too abstract to understand. """



from abc import ABC, abstractmethod
import copy

from iqmotion.communication.circular_queue import CircularQueue


class PacketState(ABC):
    """ PacketState is an abstract class that defines the state of a packet while parsing using a State Pattern. 
    It enables to parse the packet at the current state/position and find the next state depending on the current one
    until it eventually finds a whole packet and saves the message from the packet in its @property message

    It is the job of the implimentations to decide how to parse and find the next state
    """

    @abstractmethod
    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        self._byte_queue = byte_queue
        self._start_index = start_index
        self._parse_index = parse_index
        self._packet_len = packet_len

        self._is_done = False
        self._succesful = False

        self._parse_succesful = False
        self._message = bytearray([])
        self._end_index = 0

    @property
    def is_done(self):
        return self._is_done

    @property
    def is_succesful(self):
        return self._succesful

    @property
    def message(self):
        return copy.deepcopy(self._message)

    @abstractmethod
    def parse(self):
        """ this property is too abstract to understand. """

    @abstractmethod
    def find_next_state(self):
        """ this property is too abstract to understand. """



from dataclasses import dataclass


@dataclass
class ClientEntryData:
    """ ClientEntryData is an interface data class for the different client entry implementations.
    They data available in a client entry should be represented as a ClientEntryData.
    It is the implementaiton's job to also implement this data class and add any needed infromation.
    """

    type_idn: bytes
    param_idn: bytes



from iqmotion.message_making.message import Message


class ClientWithEntriesMessage(Message):
    """ ClientWithEntrisMessage is an implementation of the Message class
    It specifizes how the ClientWithEntries class makes messages
    """

    def make_bytes(self):
        """ Call the correct MessageMaker to make a message

        Returns:
            bytearray: message
        """
        return self._message_maker.make()



from iqmotion.iq_devices.base_iq_module import BaseIqModule
from iqmotion.iq_devices.custom_iq_module import CustomIqModule

# Soon to be deprecated: but the original way to access motor firmware API
from iqmotion.iq_devices.speed_module import SpeedModule
from iqmotion.iq_devices.servo_module import ServoModule
from iqmotion.iq_devices.step_dir_module import StepDirModule

# Make Motor Modules front facing visible
from iqmotion.iq_devices.vertiq8108_module import Vertiq8108
from iqmotion.iq_devices.vertiq2306_module import Vertiq2306
from iqmotion.iq_devices.fortiq_module  import Fortiq

from iqmotion.communication.communicator import Communicator
from iqmotion.communication.serial_communicator import SerialCommunicator



import sys
import glob
from queue import Queue

import serial

from iqmotion.communication.communicator import Communicator
from iqmotion.communication.crc import Crc
from iqmotion.communication.serial_packet_queue import SerialPacketQueue
from iqmotion.custom_errors import CommunicationError


class SerialCommunicator(Communicator):
    """SericalCommunicator enables to create serial packets with a message inside
    as well as extract well formed messages from its serial packets

    General Packet Format:
        | 0x55 | length | message | crcL | crcH |
        'length' is the (uint8) number of bytes in 'payload'
        'message' is a series of (uint8) bytes, serialized Little-Endian
        'crc' is the (uint16) CRC value for 'length'+'type'+'payload', Little-Endian
    """

    def __init__(self, port_name: str, baudrate=115200):
        """ Creates a SericalCommunication with a given port and baudrate

        Args:
            port_name (str): serial port name
            baudrate (int): baudrate for serial communication
        """
        self._ser_handle = self._init_serial(port_name, baudrate)
        self._out_queue = Queue()
        self._in_queue = SerialPacketQueue()

    def __del__(self):
        if hasattr(self, "_ser_handle") and self._ser_handle.is_open:
            self._ser_handle.close()
            del self._ser_handle

    def _init_serial(self, port_name: str, baudrate):
        try:
            ser_handle = serial.Serial(port_name, baudrate)

        except serial.SerialException:
            available_ports = self._find_serial_port()
            pretty_available_ports_str = "\n".join(
                '\t"{}"'.format(port) for port in available_ports
            )
            raise CommunicationError(
                f"Serial port '{port_name}' is not available, here is a list of available ports:\n{pretty_available_ports_str}"
            )

        return ser_handle

    def _find_serial_port(self):
        if sys.platform.startswith("win"):
            ports = ["COM%s" % (i + 1) for i in range(256)]
        elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
            # this excludes your current terminal "/dev/tty"
            ports = glob.glob("/dev/tty[A-Za-z]*")
        elif sys.platform.startswith("darwin"):
            ports = glob.glob("/dev/tty.*")
        else:
            raise EnvironmentError("Unsupported platform")

        working_ports = []
        for port in ports:
            try:
                s = serial.Serial(port)
                s.close()
                working_ports.append(port)
            except (OSError, serial.SerialException):
                pass

        return working_ports

    def send_message(self, message: bytearray):
        """ Packages a message and sends it through serial

        Args:
            message (bytearray): message to be packaged and sent
        """

        packet = self._make_packet(message)
        self.add_to_out_queue(packet)
        self.send_now()

    def _make_packet(self, message: bytearray):
        message_length = len(message) - 1  # minus the typeidn
        packet = message.copy()
        packet.insert(0, message_length)
        crc = Crc.make_crc(packet)
        packet.extend(crc)
        packet.insert(0, 0x55)

        return packet

    def add_to_out_queue(self, out_bytes: bytearray):
        """ Add raw bytes to the out queue, call "send_now" to send

        Args:
            bytes (bytearray): raw bytes to send
        """
        self._out_queue.put(out_bytes)

    def send_now(self):
        """ Sends everything that was added to the out queue
        """
        all_messages = bytearray([])
        while not self._out_queue.empty():
            message = self._out_queue.get()
            all_messages.extend(message)

        self._ser_handle.write(all_messages)

    def flush_input_buffer(self):
        """Flushes all the bytes left in the serial buffer in
        """
        self._in_queue.clear()
        self._ser_handle.reset_input_buffer()

    def read_bytes(self):
        """ Read bytes available in the serial port and puts them inside the packet queue.

        If there are more bytes ready to be read from serial than space in the packet queue, 
        only read up to the max packet queue size
        
        Returns:
            bool -- True if every bytes were read from the serial buffer, False otherwise
        """
        bytes_ready = self._ser_handle.in_waiting
        free_space = self._in_queue.free_space

        if bytes_ready > free_space:
            bytes_read = self._ser_handle.read(free_space)
            every_bytes_read = 0

        else:
            bytes_read = self._ser_handle.read(bytes_ready)
            every_bytes_read = 1

        self._in_queue.put_bytes(bytes_read)

        return every_bytes_read

    @property
    def bytes_left_in_queue(self):
        """ Returns the amount of bytes left in the packet queue
        """
        if self._in_queue.is_empty:
            return False

        return True

    def extract_message(self):
        """ Extract a well form message from the packet queue

        Returns:
            message (bytearray): if parsing successfull
            message (None): if not message is available
        """
        if self.bytes_left_in_queue:
            message = self._in_queue.peek()
            self._in_queue.drop_packet()
        else:
            message = None

        return message



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="iqmotion",
    packages=find_packages(),
    include_package_data=True,
    version="0.11.6",
    license="lgpl-3.0",
    description="Python libraries to talk to IQ Motion Control devices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Raphael Van Hoffelen",
    author_email="raf@iq-control.com",
    url="https://github.com/iq-motion-control/iq-module-communication-python",
    download_url="https://github.com/iq-motion-control/iq-module-communication-python/releases",
    keywords=[
        "IQ",
        "IQ CONTROL",
        "IQ MOTION CONTROL",
        "API",
        "IQ MODULES",
        "IQ LIBRARIES",
    ],
    install_requires=["numpy", "pyserial"],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.7",
)



from iqmotion.communication.packet_queue import PacketQueue
from iqmotion.communication.circular_queue import CircularQueue
from iqmotion.communication.packet_parser import PacketParser
from iqmotion.communication.serial_packet_states import SerialStartState
from iqmotion.custom_errors import PacketQueueError


class SerialPacketQueue(PacketQueue):
    """ SerialPacketQueue enables extraction of well formed, crc-verified packets from its circular byte queue

    It is a specialized circular queue which takes in raw bytes and returns the packet's message.
    The packet's message is a byte array consisting of a type byte followed by payload bytes.

    General Packet Format:
        | 0x55 | length | type | ---payload--- | crcL | crcH |
        'length' is the (uint8) number of bytes in 'data'
        'type' is the (uint8) message type
        'payload' is a series of (uint8) bytes, serialized Little-Endian
        'crc' is the (uint16) CRC value for 'length'+'type'+'data', Little-Endian
    """

    # Don't forget to change the __init__ descriptiton
    _MAX_BYTE_QUEUE_SIZE = (255 + 5) * 2  # size of max packet * 2

    def __init__(self):
        """ Create an empty SerialPacketQueue of constant size 512
        """

        self._byte_queue = CircularQueue(maxlen=self._MAX_BYTE_QUEUE_SIZE)
        self._has_a_packet = 0

    def __len__(self):
        return len(self._byte_queue)

    @property
    def is_empty(self):
        """ Check if queue is empty

        Returns:
            true: if the queue is empty
            false: if the queue has some bytes in it
        """
        return self._byte_queue.is_empty

    @property
    def free_space(self):
        current_size = len(self._byte_queue)
        return self._byte_queue.maxlen - current_size

    def put_bytes(self, new_bytes):
        """ Append bytes to the queue

        Args:
            new_bytes: bytes you want to append.

        Returns:
            true: if succesful

        Raise:
            PacketQueueError("Byte Queue Overflow")
        """
        try:
            iter(new_bytes)
        except TypeError:
            # not iterable
            if not self._byte_queue.append(new_bytes):
                raise PacketQueueError("Byte Queue Overflow")
        else:
            # iterable
            if not self._byte_queue.extend(new_bytes):
                raise PacketQueueError("Byte Queue Overflow")

        return 1

    def peek(self):
        """ Peeks for the first available packet in the queue and extract its message

        Returns:
            bytearray: The first available message
            None: There was no well form packet in the queue

        Raise:
        When the queue is empty
            PacketQueueError("Serial packet queue is empty")
        """
        if self._byte_queue.is_empty:
            raise PacketQueueError("Serial packet queue is empty")

        return self._parse_bytes()

    def _parse_bytes(self):
        packet_state = SerialStartState(self._byte_queue)

        packet_parser = PacketParser(packet_state)
        packet_parser.parse()

        if not packet_parser.succesful:
            self._has_a_packet = 0
            return None

        self._has_a_packet = 1
        return packet_parser.message

    def clear(self):
        """ Clears the queue """
        self._byte_queue.clear()

    def drop_packet(self):
        """ Drops the first packet available in the queue

        Returns:
            True: a packet was droped
            False: no packet was droped
        """
        # no need to parse if you know there is a packet
        if self._has_a_packet:
            self._pop_packet_bytes()
            self._has_a_packet = 0
            return True

        self._parse_bytes()
        if self._has_a_packet:
            self._pop_packet_bytes()
            self._has_a_packet = 0
            return True

        return False

    def _pop_packet_bytes(self):
        msg_len = self._byte_queue[1]

        # + start byte, len, type, crch, crcl
        packet_len = msg_len + 5

        for _ in range(packet_len):
            self._byte_queue.popleft()



from abc import ABC, abstractmethod


class PacketQueue(ABC):
    """ PacketQueue is an interface class that enables extraction of well formed message from its circular byte queue

    It is a specialized circular queue which takes in raw bytes and returns the packet's message.
    The packet's message is a byte array consisting of a type byte followed by payload bytes.

    It is the implementation's job to parse the packet correctly (for that implementation) and return the packet's message"""

    @abstractmethod
    def put_bytes(self, new_bytes):
        """ this property is too abstract to understand. """

    @abstractmethod
    def peek(self):
        """ this property is too abstract to understand. """

    @abstractmethod
    def drop_packet(self):
        """ this property is too abstract to understand. """
